# Sari Khaleel  sari.khalil@gmail.com
# Last modified 2-13-2011 12.00 pm

#!/usr/bin/perl
#use strict;
use POSIX;
use FindBin;
use Getopt::Long;

my $program=`basename $0`;
chomp $program;
use strict;
use warnings;

my $version = 1.0;

my $usage =<<USAGE;
USAGE:

'perl $program -h'    for this message
'perl $program -r'    for this message + the detailed README.txt file

FOR A SINGLE FILE:
  perl $program -f contigs_fasta_file [-m n]

FOR MULTIPLE FILES:
 Full report:
  perl $program -i input_dir [-s subDir1__subDir2_..] -K n_m_l -p filename_pattern
               [-c -g (k/s) -m INT -altK STRING -o]
 
 Single-stat, csv-formatted report for Excel:
  perl $program -i input_dir [-s subDir1__subDir2_..] [-K n_m_l] -p filename_pattern
                -v stat_name [-m INT -altK STRING -o]

OPTIONS:
-m INT               : Min length for Contig to be included. Default = 200.
-p filename_pattern  : the pattern of assembly filename in the directory. The 'K' cha-
                       racter (if it's in it) is replaced by K-mer lengths listed in
											 -K. See README.txt for examples
-s subDir1__subDir2..: the sub-directories within the main input directory, joined by '__'
-K n_m_l_..          : n_m_l are INTEGER K-mer lengths. They replace the 'K' character in
                       filename_pattern by default. NOTE THAT they are joined by a SINGLE
			           '_' character, while subDirs are joined by 
					   TWO '_' characters.
-c                   : print output in csv format
-g (k/s)             : group assembly files by (k)mer length or (s)ubdir. Default is s 
-altK STRING         : specify an alternative pattern of character(s) to K in case your
                       normal filename pattern has K. See README.txt for examples
-o                   : over-write previously created individual reports (see README.txt).
-v stat_name         : single_stat name (see below). 

-v: ALLOWED STAT NAMES 
----------------------
	n25, n50, n50p, n75, n90, median, mean, min, max, total, n, gc_c, gc_p
	
	n50p    : N50(perc_contigs_before, perc_contigs_after)
	median  : median contig length
	mean    : mean contig length
	min     : min contig length
	max     : max contig length
	total   : total contig length
	n       : number of contigs
	gc_c    : GC count
	gc_p    : GC%

USAGE

#0000 : Declare the global variables
my $bin_length = 10000;

my ($single_file, $inputDir, $pattern, $Ks_string, @Ks, $subDirs_string,@subDirs,
		$help, $csv, $delimiter, $overwrite, $readme);
my $min_contig_length = 200;
my $stat_name = "n50"; my $opt_stat_name;
my $exp_len = 1;
my $altK = "K";
my $groupBy = "s";
my @stat_names = qw (total n max mean median min n25 n50 n50p n75 n90 gc_c gc_p);
my $local_report_filename;
my $emergency_report; # this is used ONLY if the create/get local report action fails to work
# -------------------------------------------------------------------
#0010 : Handle the commandline arguments, and create the report header
# -------------------------------------------------------------------
&handle_commandline;

 # ReportHeader
my $report_header = "Total_length_of_sequences_(bp)$delimiter".
	"Total_number_of_contigs$delimiter"."Max_contig_length_(bp)$delimiter"."Mean_contig_length_(bp)$delimiter".
  "Median_contig_length_(bp)$delimiter"."Min_contig_length_(bp)$delimiter"."N25_(bp)$delimiter"."N50_(bp)$delimiter".
  "N50(perc_contigs_before : perc_contigs_after)_(bp)$delimiter"."N75_(bp)$delimiter"."N90_(bp)$delimiter".
	"Total_GC_count_(bp)$delimiter"."GC%"."\n";

$report_header = "Experiments$delimiter"."K-mer_size_(bp)$delimiter".$report_header if !$single_file;

# -------------------------------------------------------------------	
#0020 : See if it's just a single file report
# -------------------------------------------------------------------
# Get the single file report
  &print_single_file_report if ($single_file); # Get the report and exit

# -------------------------------------------------------------------
# 0030 Unless it's a single_file report,
#      Extract data from pre-existing files (unless overwrite is ON)
#      AND calculate progress
# -------------------------------------------------------------------
	print "Calculating. This can take a while ..\n";
	my %exp_k_hash;
	my ($total_assemblies_c, $finished_assemblies_c)  = (0,0);
	
	&get_and_create_ALL_reports();

# -------------------------------------------------------------------
# 0040 Print progress, date
# 	   [AND print stats/single_stat_table]
# -------------------------------------------------------------------

my $date =`date`; $date =~ s/\n//;
if ($total_assemblies_c){
	my $rounded = sprintf("%.4f", ($finished_assemblies_c/$total_assemblies_c));
	print "\nFound: ".($rounded * 100)."% ($finished_assemblies_c/$total_assemblies_c) of the expected assembly files at $date\n\n";
}
else {
	die "ERROR: no assembly files were detected. Your path is probably wrong\n";
}

if ($opt_stat_name){ &print_single_stat(\%exp_k_hash, $stat_name, \@Ks);}
else           { &print_stats(\%exp_k_hash);}

if ($emergency_report){
	print "\nThere were errors creating/getting the local report at some Kmer directories.\nHowever, I managed to save the following reports:\n";
	print $report_header;
	print $emergency_report."\n";
}
exit;
####################################################################################





############################# SUBROUTINES ##########################################

sub handle_commandline{
	
	die $usage unless @ARGV;
	
	GetOptions(
			 "file=s" => \$single_file,
			 "inputDir=s" => \$inputDir,
			 "subDirs=s" => \$subDirs_string,
			 "Ks=s" => \$Ks_string,
			 "pattern=s" => \$pattern,
			 "groupBy=s" => \$groupBy,
			 "csv" => \$csv,
			 "min_contig_length=i" => \$min_contig_length,
			 "help" => \$help,
			 "readme"=> \$readme,
	#	   "exp_len=i" => \$exp_len,
			 "altK=s" => \$altK,
			 "v=s" => \$opt_stat_name,
			 "overwrite"=>\$overwrite
			 );
	
		if ($opt_stat_name) {$stat_name = $opt_stat_name;}
		
		# Handle help
		die $usage if $help ;
		if ($readme){
			system ("cat $FindBin::Bin/README.txt");
			exit;
		}
	
	# Set the Delimiter
		if ($csv || $opt_stat_name) {$delimiter = ",";}
		else {$delimiter = "\t";}
	
	# Checks for inputDirs
		if (! $single_file){
			&check_vars; 

			# Extract @Ks, @subDirs
			if ($Ks_string) { @Ks = split (/_/, $Ks_string); } else {push @Ks, "";}
			if (@Ks) {print "Ks: @Ks\n";}
			
			if ($subDirs_string) {@subDirs = split (/__/, $subDirs_string);} else {push @subDirs, "";}
			if (@subDirs) {print "subDirs: @subDirs\n";}
			
		 if (($groupBy !~ /s/i) && ($groupBy !~ /k/i)){
			 die "ERROR: invalid value for groupBy: $groupBy, must be either 'subdir', 's', or 'kmer', 'k' (case insensitive) !!\n\n";
		 }
		}# end if ! single_file
		
}# end sub handle commandline

sub check_vars{ 
		# Handle missing essential arguments
		if (! $inputDir) {die "ERROR: No input dir (-i STRING) was provided !!\n\n";}
		if (! $pattern)  {die "ERROR: No filename pattern (-p STRING) was provided !! \n\n";}
		if (! $Ks_string)  {die "ERROR: No list of Kmers (-K n_m_l) was provided !! \n\n";}
	
		#Handle wrong arguments
			#Ks doesn't have a number
			if ($Ks_string) {
				#print $Ks_string."\n";
				if ($Ks_string !~ /\d+/){ die "Ks did not include any numbers !\n\n";}}
			#Ks doesn't have double underscore joins
			#if ($Ks_string) {
				#print $Ks_string."\n";
				#if ($Ks_string !~ /__/){ die "The values in Ks are not joined by double underscores '__' !\n\n";}}
			
			
			#inputDir does not exist or is not a directory
			if (! -d $inputDir) {die "inputDir, $inputDir, does not exist or is not a directory !\n\n";}
			
			#Check subDirs
			if ($subDirs_string){
				#subDirs doesn't have double underscore joins
				#if ($subDirs_string !~ /__/){ die "The values in subDirs are not joined by double underscores '__' !\n\n";}
	
				#one of the subDirs do not exist
				my @check_subDirs = split (/__/, $subDirs_string);
				foreach (@check_subDirs){
					my $check_subDir = $inputDir."/".$_;
					if (! -d $check_subDir){
						print "The following subDir, $check_subDir, is not a directory or does not exist !! \n\n";	
					} 
				}	
			}
			
			#groupBy is neither Kmer nor experiment
			if ($groupBy){
				if (($groupBy !~ /s/i) && ($groupBy !~ /k/i)){
					die "invalid value for groupBy: $groupBy, must be either 'subdir', 's', or 'kmer', 'k' (case insensitive) !!\n\n";
				}
			}
		
		# Handle wrong argument combinations
			#Ks or subDirs, but no groupBy
			if (($Ks_string || $subDirs_string) && (! $groupBy)) {die "Multiple subdirs and/or Ks were provided, but no grouping option was given !! \n\n"; }
			#groupBy, but neither Ks nor subDirs
		if ( (! ($Ks_string || $subDirs_string)) && ($groupBy)) {die "a grouping option was given, but there are no multiple Ks or subDirs !! \n\n"; }

}#end sub check_vars

sub print_single_file_report{
	die "This file, $single_file, does not exist !!\n" if (! -e $single_file);
	my $single_report = getStatistics($single_file, $delimiter);
	print $report_header.$single_report if $single_report;
	exit;
}



sub get_string_index {
	# Usage: returns the index of a query string item in an array, -1 if not found
	#         Matching is either case sensitive or not 
	
	my ($arr_ref, $query, $case_sensitive) = @_;
	#die "@$arr_ref \n$query\n$case_sensitive\n";
	my $index = 0;
	
	if ($case_sensitive){
		while (($$arr_ref[$index] ne $query) && ($index <= $#$arr_ref)){
			$index++;
		} 
	}
	else{
		while ((($query !~ /$$arr_ref[$index]/i) || ((length $query) != (length $$arr_ref[$index]))) && ($index <= $#$arr_ref)){
			$index++;
		}
	}# end else
	
	return "-1" if ($index > $#$arr_ref);
	return $index;
}# end sub get string index

sub create_local_report {
	# Usage: Creates a local report file at the kmer_dir
	#        The report's contents vary depending on the assembler and the used genome
	my ($subDir, $dir_of_contigs_file, $local_report_filename, $contigs_file, $kmer) = @_;
	
	# If subDir is defined, the contigs file path is:
	# inputDir/subDir/_K/contigs_filename. $exp = subDir
	my $exp;
	if ($subDir){
		my @subDir_path_parts = split /\//, $subDir;
		$exp = $subDir_path_parts[$#subDir_path_parts];
	}
	else {
		# if subDir is not defined, this means that the user did not use -s and there are no subDirs
		# the contigs filepath is inputDir//_K/contigs_filename. $exp = inputDir
		my @contigs_filepath_parts = split /\//, $contigs_file;
		$exp = $contigs_filepath_parts[-4];
	}
	
	open OUT, "> $dir_of_contigs_file/$local_report_filename";
	
	my $report_line = "$exp,$kmer,".&getStatistics($contigs_file,",");
	
	print OUT $report_header.$report_line;
	close OUT;
	#die "$dir_of_contigs_file/$local_report_filename";
}# end sub create local report

sub get_and_create_ALL_reports{
	$local_report_filename = "indiv_assembly_report_minContigLength_$min_contig_length"."_v.$version.txt"; 
	
	foreach my $kmer (@Ks){
		foreach my $subDir(@subDirs){
				# Create the filename from the pattern
				my $filename = $pattern;
				$filename =~ s/$altK/$kmer/g; 
				my @contig_filepath_parts = split /\//, $filename;
				my $dir_of_contigs_file = $contig_filepath_parts[$#contig_filepath_parts - 1]; 
				
				$total_assemblies_c++;
				my $contigs_file = "$inputDir/$subDir/$filename";
				$dir_of_contigs_file = "$inputDir/$subDir/$dir_of_contigs_file";
				print "Getting assembly_report for $contigs_file ..\n";
				if (-e $contigs_file) {
					$finished_assemblies_c++;
					my $pre_existing_report = 0;
					if (! $overwrite){
						$pre_existing_report = &get_local_report(\%exp_k_hash, $dir_of_contigs_file, $delimiter);				
					}
					# If there's no pre-existing report* AND/OR it's not just progress
					# * or if the pre-existing report is to be overwritten
					if (!$pre_existing_report){
						&create_local_report($subDir, $dir_of_contigs_file, $local_report_filename,$contigs_file, $kmer);
						$pre_existing_report = &get_local_report(\%exp_k_hash, $dir_of_contigs_file, $delimiter);
						
						if (!$pre_existing_report){
							print "ERROR: failed to create and get individual report at $dir_of_contigs_file!\n" ;
							$emergency_report .= $subDir.$delimiter.$kmer.$delimiter.&getStatistics($contigs_file,$delimiter);
						}
					}
				}# end if (-e $contigs_file)
				else {
						print "WARNING: contigs file not found at $contigs_file  !!\n";
				}
		}#end foreach subdirs
	}# end foreach (@Ks)
}# end sub get_and_create_ALL_reports

sub get_local_report{
	# Usage: For a specific K-mer assembly dir, it tries to parse a pre-existing report
	#        for the experminent's name and the rest of the line, then fills exp_k_hash{exp_K}
	#        with the rest of the line.
	
	my ($exp_k_hash_ref, $dir_of_contigs_file, $delimiter) = @_;
	
	if (-e "$dir_of_contigs_file/$local_report_filename"){
		open IN, "< $dir_of_contigs_file/$local_report_filename";
		my $line = <IN>; # skip the header line
		while ($line = <IN>){
			if ($line =~ /Exp_\w+,\d\d/){
				#print $line;
				my $exp_k = $&;
				my $rest_of_line = $';
				
				if ($rest_of_line =~ /\d+/){
					$exp_k =~ s/,/$delimiter/;
					$rest_of_line =~ s/,/$delimiter/g;
					#die "->$rest_of_line<-$delimiter!";
					$$exp_k_hash_ref{$exp_k} = $rest_of_line;
					#print "$exp_k = $rest_of_line\n";
					#print "$exp_k = $$exp_k_hash_ref{$exp_k}\n";
				}
				else {print "ERROR: no numbers were detected in this experiment line : $line\n"; return 0;}
			}
			else {print "ERROR: Experiment line is not normal: $line\n"; return 0;}
		}# end while
	return 1;
	}# end if (-e previous report)
	else {# if ! -e previous report
		return 0;
	}
}# end sub get_local_report

sub print_stats{
	# Usage : prints a delimiter-delimited table of exp-K, stats grouped by Kmer or subDir
	          
	my ($exp_k_hash_ref) = @_;
	
	print $report_header;
	
	if (! @subDirs) {
		$subDirs[0] = "";
	}
	if (! @Ks) {
	   $Ks[0] = "";  
	}
		
	if ($groupBy =~ /k/i){	
		foreach my $k(@Ks){
			foreach my $subDir (@subDirs){
					my $exp;
					if ($subDir){
						my @subDir_path_parts = split /\//, $subDir;
						$exp = $subDir_path_parts[$#subDir_path_parts];
					}
					else {# if subDir is not defined because the user did not use -s
					    # then exp = the inputDir itself
						my @inputDir_parts = split /\//, $inputDir;
						$exp = $inputDir_parts[-1];
					}
					if ($$exp_k_hash_ref{"$exp$delimiter$k"}){
						print "$exp$delimiter$k".$$exp_k_hash_ref{"$exp$delimiter$k"};
					}
			}#end foreach subdirs
		}# end foreach (@Ks)
	}#end if groupBy =~ k
	
	elsif ($groupBy =~ /s/i){
		foreach my $subDir (@subDirs){
				my $exp;
				if ($subDir){
					my @subDir_path_parts = split /\//, $subDir;
					$exp = $subDir_path_parts[$#subDir_path_parts];
				}
				else {# if subDir is not defined because the user did not use -s
					    # then exp = the inputDir itself
					my @inputDir_parts = split /\//, $inputDir;
					$exp = $inputDir_parts[-1];
				}
				foreach my $k(@Ks){
					if ($$exp_k_hash_ref{"$exp$delimiter$k"}){
						print "$exp$delimiter$k".$$exp_k_hash_ref{"$exp$delimiter$k"};
					}
				}# end foreach (@Ks)
		}#end foreach subdirs
	}# end if groupBy subdir
}# end sub print_stats

sub print_single_stat{
	# Usage: Parses exp_k_hash (which was filled by get_local_report) to extract the values of
	#        a single stat and print them in a clustered excel table format.
	
	my ($exp_k_hash_ref, $stat) = @_;
	
	my $stat_index = &get_string_index(\@stat_names, $stat, 1);
	if ($stat_index == -1) {die "ERROR: invalid stat_name ($stat)!\nValid stat names: @stat_names\n";}
	my $report_header_copy = $report_header;
	chomp ($report_header_copy);
	my @stat_titles = split /,/, $report_header_copy;
	
	my $kmers_string ="";
	foreach (@Ks){
		$kmers_string .= ",_".$_;
	}
	
	print "STAT: $stat [ ".$stat_titles[$stat_index+2]." ]\n\n";
	print "Experiments,K-mer lengths (bp)\n";
	print "$kmers_string\n";

	foreach my $subDir (@subDirs){
				my $exp;
				if ($subDir){
					my @subDir_path_parts = split /\//, $subDir;
					$exp = $subDir_path_parts[$#subDir_path_parts];
				}
				else {# if subDir is not defined because the user did not use -s
					    # then exp = the inputDir itself
					my @inputDir_parts = split /\//, $inputDir;
					$exp = $inputDir_parts[-1];
				}
			print "$exp";	
			foreach my $k(@Ks){
				if ($$exp_k_hash_ref{"$exp$delimiter$k"}){
					my @stats_array = split /$delimiter/, $$exp_k_hash_ref{"$exp$delimiter$k"};
					#print @stats_array." @stats_array\n";
					my $stat_value = $stats_array[($stat_index+1)];
					chomp ($stat_value);
					if ($stat eq "n50"){
						if ($stat_value =~ /^\d+/){$stat_value = $&;}
					}
					
					print ",$stat_value";
				}
			}# end foreach (@Ks)
			print "\n";
	}#end foreach subdirs
}# end print_single_stat


# ============================ GET STATISTICS SUBROUTINES ==============================

sub getStatistics {
	my ($contigsFile, $delimiter) = @_;
	my $Id;
	my %seq_len;
	my %seq_gc;
	my $report = "";
	#print $contigsFile."\n";
	if(-e $contigsFile) {
		my $file_has_at_least_one_contig;
		open(CONTIGS, $contigsFile) or die"Can't open file $contigsFile\n";
		while (my $line=<CONTIGS>) {
			chomp($line);
			if ($line =~ /^\>/)  { 
				$Id = $line; 
			}
			else 	{ 
				if ($Id){
					$seq_len{$Id} += length ($line); $file_has_at_least_one_contig = 1 if $seq_len{$Id};
					$seq_gc{$Id} += ($line =~ tr/gGcC/gGcC/);	
				}
			}
		}
		close (CONTIGS);
		if (!$file_has_at_least_one_contig){
			print "ERROR: zero sequences detected in $contigsFile !!\n\n";
		}
		else{
			# Count the number of sequences in the file and create a histogram of the distribution
			my $n = 0;
			my $int;
			my $totalLength = 0;
			my $gcCount = 0;
			my %len= ();
			my @seqLengths;
			foreach my $id (keys %seq_len) {
				if ($seq_len{$id} >= $min_contig_length){
					push (@seqLengths, $seq_len{$id}); # record length for N50 calc's
					$n++;
					$int = floor( $seq_len{$id}/$bin_length );
					$totalLength += $seq_len{$id};
					$gcCount += $seq_gc{$id};
					if( !defined($len{$int}) ) {
						$len{$int} = 1;
					} else {
						$len{$int}++;
					}
				
				}#end if >= min contig length
			}#end foreach
		
			# Calculate min, max, median
			my $minContigLength = min(\@seqLengths);
			my $maxContigLength = max(\@seqLengths);
			my $medianContigLength = median(\@seqLengths);
			my $averageContigLength = average(\@seqLengths);
		
			# Calculate N25, N50, and N75 and counts
			my $N25; my $N50; my $N75;
			my $N25count=0; my $N50count=0; my $N75count=0;
			my $frac_covered = $totalLength;
			@seqLengths = reverse sort { $a <=> $b } @seqLengths;
			$N25 = $seqLengths[0];
			while ($frac_covered > $totalLength*3/4) {
				$N25 = shift(@seqLengths);
				$N25count++; $N50count++; $N75count++;
				$frac_covered -= $N25;
			}
			$N50 = $N25;
			while ($frac_covered > $totalLength/2) {
				$N50 = shift(@seqLengths);
				$N50count++; $N75count++;
				$frac_covered -= $N50;
			}
			$N75 = $N50;
			while ($frac_covered > $totalLength/4) {
				$N75 = shift(@seqLengths);
				$N75count++;
				$frac_covered -= $N75;
			}
			my $N90 = $N75; my $N90count = 0;
			while ($frac_covered > $totalLength/10) {
				$N90 = shift(@seqLengths);
				$N90count++;
				$frac_covered -= $N90;
			}
		
		
			my $perc_after = $N50count/(keys %seq_len);
			$perc_after = sprintf("%.3f", $perc_after);
			my $perc_before = 100 - $perc_after;
			
			# Print out the results
			$report .="$totalLength$delimiter";
			$report .= "$n$delimiter";
			$report .= "$maxContigLength$delimiter";
			$report .=sprintf("%.2f", $averageContigLength)."$delimiter";
			$report .= "$medianContigLength$delimiter";
			$report .= "$minContigLength$delimiter";
			$report .="$N25$delimiter";
			$report .="$N50$delimiter";
			$report .="$N50($perc_before"."% : $perc_after"."%)$delimiter";
			$report .="$N75$delimiter";	
			$report .="$N90$delimiter";	
			$report .="$gcCount$delimiter";
			$report .= sprintf("%.3f",($gcCount/$totalLength * 100))."\n";		
		}# end else (if file_has_at_least_one_contig)
	}# end if (-e $contigsFile)
	return $report;
}

sub min {
	@_ == 1 or die ('Sub usage: $min = min(\@array);'); 
	my ($array_ref) = @_;
        my @array = sort { $a <=> $b } @$array_ref;
	#print "@array\n";
	return $array[0]; 
}

sub max {
	@_ == 1 or die ('Sub usage: $max = max(\@array);'); 
	my ($array_ref) = @_;
        my @array = sort { $a <=> $b } @$array_ref;
	my $count = scalar @$array_ref;
	return $array[$count-1]; 
}

sub average() {
 	@_ == 1 or die ('Sub usage: $average = average(\@array);');
	my ($array_ref) = @_;
	my $sum =0;
	my $count = scalar @$array_ref;
	foreach (@$array_ref) { 
		$sum += $_; 
	}
	return $sum / $count;
}

sub median() {
	@_ == 1 or die ('Sub usage: $median = median(\@array);');
	my ($array_ref) = @_;
	my $count = scalar @$array_ref;
	# Sort a COPY of the array, leaving the original untouched
	my @array = sort { $a <=> $b } @$array_ref;
	if ($count % 2) {
		return $array[int($count/2)];
	} else {
		return ($array[$count/2] + $array[$count/2 - 1]) / 2;
	}
} 
